// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CONCEPT_TOOLS_HH
# define CONCEPT_TOOLS_HH

# include <type_traits>
# include <memory>
# include <cassert>
# include "../wrappers/operators.hh"
# include "parse_ids.hh"

/// \defgroup Macros Macro definitions
/// \{

/// \brief Enable function definition if concept is fulfilled
///
/// It should be used in the parameter list a function template.
# define ENABLE_IF(X...) \
  typename = typename std::enable_if< ::catsfoot::eval< X >::value>::type

/// \brief Enable function definition if concept has no model
///
/// It should be used in the parameter list a function template.
# define ENABLE_IF_NOT(X...) \
  typename = typename std::enable_if<!::catsfoot::eval< X >::value>::type

/// \brief Tells whether a concept is fulfilled.
///
/// It makes a constant expression of type \a bool.
# define IF(X...)				\
  ::catsfoot::eval< X >::value

/// \brief Defines a list of axiom in a concept
# define AXIOMS(X...)						\
  static auto get_axioms() ->					\
    decltype(::catsfoot::details::zip_vec_tuple			\
	     (::catsfoot::details::split_identifiers(#X),	\
	      std::make_tuple(X))) {				\
    return ::catsfoot::details::zip_vec_tuple			\
      (::catsfoot::details::split_identifiers(#X),		\
       std::make_tuple(X));					\
  }

/// \}

#include "has_requirements.hh"
#include "is_concept.hh"
#include "static_and.hh"

namespace catsfoot {
  /// \brief User type traits for validating models.
  /// \ingroup UserTraits
  template <typename T>
  struct verified: public std::false_type {
  };

  /// \brief List representation of several concepts
  template <typename... T>
  struct concept_list {};

  namespace details {
    /// \brief Default case: predicate
    template <typename T,
	      bool = has_requirements<T>::value,
	      bool = is_concept<T>::value,
	      bool = is_auto_concept<T>::value>
    struct eval: public T {
    };

    /// \brief If T is a concept model
    template <typename T, bool B>
    struct eval<T, true, true, B>:
      public static_and<eval<typename T::requirements>,
			verified<T> > {
    };

    /// \brief If T is an auto-concept model
    template <typename T>
    struct eval<T, true, false, true>:
      public static_and<eval<typename T::requirements>>
    {
    };

    /// \brief If the parameter is a list of requirements
    template <typename... T>
    struct eval<concept_list<T...>, false, false, false>:
      public static_and<eval<T>... > {
    };
  }

  /// \brief Predicate evaluating models and predicates.
  /// \ingroup Predicates
  template <typename T>
  struct eval: public details::eval<T> {
  };

  namespace details {
    /// \brief Check that T is a verified model.
    template <typename T>
    struct class_assert_verified {
      static_assert(verified<T>::value,
		    "Concept was not verified");
    };

    /// \brief Concept checking: predicate
    template <typename T,
	      bool = is_concept<T>::value,
	      bool = is_auto_concept<T>::value,
	      bool = has_requirements<T>::value>
    struct class_assert_concept {
      static_assert(T::value, "Missing requirement");
    };

    /// \brief Concept checking: recursion on a list of requirements
    template <typename F, typename... T, bool A, bool B, bool C>
    struct class_assert_concept<concept_list<F, T...>, A, B, C>:
      public virtual class_assert_concept<F>,
      public virtual class_assert_concept<concept_list<T...> > {
    };

    /// \brief Concept checking: empty list of requirements
    template <bool A, bool B, bool C>
    struct class_assert_concept<concept_list<>, A, B, C> {
    };

    /// \brief Concept checking: T is a concept model
    template <typename T, bool B>
    struct class_assert_concept<T, true, B, true>:
      public virtual class_assert_concept<typename T::requirements>,
      public virtual class_assert_verified<T> {
    };

    /// \brief Concept checking: T is a auto-concept model
    template <typename T>
    struct class_assert_concept<T, false, true, true>:
      public class_assert_concept<typename T::requirements> {
    };

  }

  /// \brief Checks that a concept is
  ///
  /// This type will raise a static assertion on when instantiating its
  /// concrete type if T is not a valid concept. It is intended to be
  /// used in a class template. For example:
  /// \code
  /// template <typename T>
  /// struct foo {
  ///   //...
  ///   class_assert_verified<some_concept<T> > check;
  /// };
  /// \endcode
  template <typename T>
  struct class_assert_concept:
    public details::class_assert_concept<T> {
  };


  /// \brief Assert that the concept T holds
  ///
  /// This function does not do anything. It will however raise a static
  /// assertion in case T is not a valid concept.
  template <typename T>
  void assert_concept(const T&) {
    class_assert_concept<T>();
  }

}

#endif
