// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef RANDOM_TERM_GENERATOR_HH
# define RANDOM_TERM_GENERATOR_HH

# include "../utils/call_with.hh"
# include <tuple>
# include <functional>
# include <vector>
# include <cassert>
# include <random>
# include <type_traits>
# include "../type_traits/try_first.hh"
# include "../wrappers/function_wrappers.hh"
# include "position.hh"
# include "dataset.hh"

namespace catsfoot {
  namespace details {
    template <typename T, typename... Functions>
    struct number_function_returns;

    /// \brief No function given
    template <typename T>
    struct number_function_returns<T> {
      static const size_t value = 0u;
    };

    /// \brief The first function returns a T
    template <typename T, typename... Args, typename... Functions>
    struct number_function_returns<T, std::function<T(Args...)>,
				   Functions...> {
      static const size_t value =
	number_function_returns<T, Functions...>::value + 1;
    };

    /// \brief The first function returns a T
    template <typename T, typename... Args, typename... Functions>
    struct number_function_returns<T, std::function<const T&(Args...)>,
				   Functions...> {
      static const size_t value =
	number_function_returns<T, Functions...>::value + 1;
    };

    /// \brief The first function returns a T
    template <typename T, typename... Args, typename... Functions>
    struct number_function_returns<T, std::function<T&(Args...)>,
				   Functions...> {
      static const size_t value =
	number_function_returns<T, Functions...>::value + 1;
    };

    /// \brief The first function returns a T
    template <typename T, typename... Args, typename... Functions>
    struct number_function_returns<T, std::function<T&&(Args...)>,
				   Functions...> {
      static const size_t value =
	number_function_returns<T, Functions...>::value + 1;
    };

    /// \brief The first function does not return a T
    template <typename T, typename Ret, typename... Args,
	      typename... Functions>
    struct number_function_returns<T, std::function<Ret(Args...)>,
				   Functions...> {
      static const size_t value =
	number_function_returns<T, Functions...>::value;
    };

    template <typename T, typename... Functions>
    struct number_ground_terms;

    /// \brief No function
    template <typename T>
    struct number_ground_terms<T> {
      static const size_t value = 0u;
    };

    /// \brief First function is ground term
    template <typename T, typename... Functions>
    struct number_ground_terms<T, std::function<T()>,
				   Functions...> {
      static const size_t value =
	number_ground_terms<T, Functions...>::value + 1;
    };

    /// \brief First function is ground term
    template <typename T, typename... Functions>
    struct number_ground_terms<T, std::function<const T&()>,
				   Functions...> {
      static const size_t value =
	number_ground_terms<T, Functions...>::value + 1;
    };

    /// \brief First function is ground term
    template <typename T, typename... Functions>
    struct number_ground_terms<T, std::function<T&()>,
				   Functions...> {
      static const size_t value =
	number_ground_terms<T, Functions...>::value + 1;
    };

    /// \brief First function is ground term
    template <typename T, typename... Functions>
    struct number_ground_terms<T, std::function<T&&()>,
				   Functions...> {
      static const size_t value =
	number_ground_terms<T, Functions...>::value + 1;
    };

    /// \brief First function is not ground term
    template <typename T, typename Ret, typename... Args,
	      typename... Functions>
    struct number_ground_terms<T, std::function<Ret(Args...)>,
				   Functions...> {
      static const size_t value =
	number_ground_terms<T, Functions...>::value;
    };
  }

  /// \brief Builds random term generators
  /// \ingroup DataSet
  /// \tparam Types The list of types it can generate
  template <typename... Types>
  struct term_generator_builder {
  private:
    size_t size;
  public:
    term_generator_builder(size_t size = 200u): size(size) {}

    template <typename Generator, typename... Functions>
    struct generator {
    private:
      size_t size;
      Generator& g;
      std::tuple<Functions...> functions;
      std::tuple<std::vector<Types>...> terms;

    public:
      template <typename... F>
      generator(size_t size, Generator& g, F&&... functions):
	size(size), g(g), functions(std::forward<F>(functions)...) {
      }

    private:
      template <typename Return>
      Return call_this_one(std::function<Return()>& f) {
	return f();
      }

      template <typename Return>
      Return&& call_this_one(std::function<Return&&()>& f) {
	return std::move(f());
      }

      template <typename Return>
      Return& call_this_one(std::function<Return&()>& f) {
	return f();
      }

      template <typename Return>
      const Return&& call_this_one(std::function<const Return&()>& f) {
	return f();
      }

      template <typename Return, typename... Args,
		typename... Bound, typename = typename
		std::enable_if<sizeof...(Bound) < sizeof...(Args)>::type>
      Return call_this_one(std::function<Return(Args...)>& f,
			   Bound&... bound...) {
	typedef typename
	  std::decay<
	    typename std::tuple_element<sizeof...(Bound),
					std::tuple<Args...> >::type
	    >::type
	  that_arg;
	std::vector<that_arg>&
	  l = std::get<details::position<that_arg, Types...>::value>(terms);

	if (l.size() == 0) {
	  return call_this_one(f, bound..., generate<that_arg>(-1));
	}
	if (std::is_same<that_arg, Return>::value) {
	  // It is not safe to insert new elements.
	  return call_this_one(f, bound...,
					 l[std::uniform_int_distribution<size_t>
					   (0, l.size()-1)(g)]);
	}
	return call_this_one(f, bound..., generate<that_arg>());
      }

      template <typename Return, typename... Args,
		typename... Bound, typename = typename
		std::enable_if<sizeof...(Bound) < sizeof...(Args)>::type>
      Return&& call_this_one(std::function<Return&&(Args...)>& f,
			   Bound&... bound...) {
	typedef typename
	  std::decay<
	    typename std::tuple_element<sizeof...(Bound),
					std::tuple<Args...> >::type
	    >::type
	  that_arg;
	std::vector<that_arg>&
	  l = std::get<details::position<that_arg, Types...>::value>(terms);

	if (l.size() == 0) {
	  return call_this_one(f, bound..., generate<that_arg>(-1));
	}
	if (std::is_same<that_arg, Return>::value) {
	  // It is not safe to insert new elements.
	  return call_this_one(f, bound...,
					 l[std::uniform_int_distribution<size_t>
					   (0, l.size()-1)(g)]);
	}
	return call_this_one(f, bound..., generate<that_arg>());
      }

      template <typename Return, typename... Args,
		typename... Bound, typename = typename
		std::enable_if<sizeof...(Bound) < sizeof...(Args)>::type>
      const Return& call_this_one(std::function<const Return&(Args...)>& f,
			   Bound&... bound...) {
	typedef typename
	  std::decay<
	    typename std::tuple_element<sizeof...(Bound),
					std::tuple<Args...> >::type
	    >::type
	  that_arg;
	std::vector<that_arg>&
	  l = std::get<details::position<that_arg, Types...>::value>(terms);

	if (l.size() == 0) {
	  return std::move(call_this_one(f, bound..., generate<that_arg>(-1)));
	}
	if (std::is_same<that_arg, Return>::value) {
	  // It is not safe to insert new elements.
	  return call_this_one(f, bound...,
			       l[std::uniform_int_distribution<size_t>
				 (0, l.size()-1)(g)]);
	}
	return call_this_one(f, bound..., generate<that_arg>());
      }

      template <typename Return, typename... Args,
		typename... Bound, typename = typename
		std::enable_if<sizeof...(Bound) < sizeof...(Args)>::type>
      Return& call_this_one(std::function<Return&(Args...)>& f,
			   Bound&... bound...) {
	typedef typename
	  std::decay<
	    typename std::tuple_element<sizeof...(Bound),
					std::tuple<Args...> >::type
	    >::type
	  that_arg;
	std::vector<that_arg>&
	  l = std::get<details::position<that_arg, Types...>::value>(terms);

	if (l.size() == 0) {
	  return call_this_one(f, bound..., generate<that_arg>(-1));
	}
	if (std::is_same<that_arg, Return>::value) {
	  // It is not safe to insert new elements.
	  return call_this_one(f, bound...,
			       l[std::uniform_int_distribution<size_t>
				 (0, l.size()-1)(g)]);
	}
	// Because we are returning a reference, it is good to say that we want to
	// force to call any function to ensure a new value.
	return call_this_one(f, bound..., generate<that_arg>(0));
      }

      template <typename Return, typename... Args, typename ...Bound,
		typename = typename
		std::enable_if<sizeof...(Bound) == sizeof...(Args)>::type>
      Return call_this_one(std::function<Return(Args...)>& f,
			     Bound&... bound) {
	return f(bound...);
      }

      template <typename Return>
      Return call(selector<Return>, size_t) {
	assert(false);
      }

      template <typename Return, typename... FS,
		typename... Args>
      Return call(selector<Return>, size_t fun_to_call,
		    std::function<Return(Args...)>& f,
		    FS... fs) {
	if (fun_to_call == 0) {
	  return call_this_one(f);
	} else {
	  return call
	    (selector<Return>(),
	     fun_to_call-1, fs...);
	}
      }

      template <typename Return, typename... FS,
		typename... Args>
      Return call(selector<Return>, size_t fun_to_call,
		  std::function<Return&(Args...)>& f,
		  FS... fs) {
	if (fun_to_call == 0) {
	  return call_this_one(f);
	} else {
	  return call
	    (selector<Return>(),
	     fun_to_call-1, fs...);
	}
      }

      template <typename Return, typename... FS,
		typename... Args>
      Return call(selector<Return>, size_t fun_to_call,
		  std::function<const Return&(Args...)>& f,
		  FS... fs) {
	if (fun_to_call == 0) {
	  return call_this_one(f);
	} else {
	  return call
	    (selector<Return>(),
	     fun_to_call-1, fs...);
	}
      }

      template <typename Return, typename... FS,
		typename... Args>
      Return call(selector<Return>, size_t fun_to_call,
		  std::function<Return&&(Args...)>& f,
		  FS... fs) {
	if (fun_to_call == 0) {
	  return std::move(call_this_one(f));
	} else {
	  return std::move(call
	    (selector<Return>(),
	     fun_to_call-1, fs...));
	}
      }

      template <typename Return,
		typename OtherReturn, typename... FS,
		typename... Args>
      Return call(selector<Return>, size_t fun_to_call,
		    std::function<OtherReturn(Args...)>&,
		    FS... fs) {
	return call
	  (selector<Return>(),
	   fun_to_call, fs...);
      }

    public:
      template <typename Return,
		bool moveable = is_constructible<Return(const Return&)>::value ||
		is_constructible<Return(Return&&)>::value>
      struct random_container {
      private:
	generator<Generator, Functions...>* upper;
      public:
	random_container(generator<Generator, Functions...>& upper)
	  : upper(&upper) {}
	struct iterator {
	private:
	  generator<Generator, Functions...>* upper;
	  size_t pos;

	public:
	  iterator(generator<Generator, Functions...>& upper,
		   size_t pos = 0u):
	    upper(&upper),
	    pos(pos) {
	    if (pos == this->upper->size)
	      return ;
	    std::vector<Return>& l =
	      std::get<details::position<Return, Types...>::value>
	      (this->upper->terms);
	    while (l.size() <= pos)
	      this->upper->template generate<Return>();
	  }
	  iterator(const iterator& other):
	    upper(other.upper),
	    pos(other.pos) {}
	  Return& operator*() const {
	    std::vector<Return>& l =
	      std::get<details::position<Return, Types...>::value>
	      (upper->terms);
	    return l[pos];
	  }
	  iterator& operator++() {
	    ++pos;
	    std::vector<Return>& l =
	      std::get<details::position<Return, Types...>::value>
	      (upper->terms);
	    while (l.size() <= pos)
	      upper->template generate<Return>();
	    return *this;
	  }
	  iterator& operator++(int) {
	    iterator ret(*this);
	    ++*this;
	    return *this;
	  }
	  bool operator==(const iterator& other) const {
	    return pos == other.pos;
	  }
	  bool operator!=(const iterator& other) const {
	    return pos != other.pos;
	  }

	  iterator& operator=(iterator&& other) {
	    std::swap(upper, other.upper);
	    std::swap(pos, other.pos);
	    return *this;
	  }

	  iterator& operator=(const iterator& other) {
	    return *this = iterator(other);
	  }
	};
	iterator begin() {
	  return iterator(*upper, 0);
	}
	iterator end() {
	  return iterator(*upper, upper->size);
	}
      };

      template <typename Return>
      struct random_container<Return, false> {
      private:
	generator<Generator, Functions...>* upper;
      public:
	random_container(generator<Generator, Functions...>& upper)
	  : upper(&upper) {}
	struct iterator {
	private:
	  generator<Generator, Functions...>* upper;
	  size_t pos;

	public:
	  iterator(generator<Generator, Functions...>& upper,
		   size_t pos = 0u):
	    upper(&upper),
	    pos(pos) {
	  }
	  iterator(const iterator& other):
	    upper(other.upper),
	    pos(other.pos) {}
	  Return& operator*() const {
	    return const_cast<iterator*>(this)->upper->template generate<Return>();
	  }
	  iterator& operator++() {
	    ++pos;
	    return *this;
	  }
	  iterator& operator++(int) {
	    iterator ret(*this);
	    ++*this;
	    return *this;
	  }
	  bool operator==(const iterator& other) const {
	    return pos == other.pos;
	  }
	  bool operator!=(const iterator& other) const {
	    return pos != other.pos;
	  }

	  iterator& operator=(iterator&& other) {
	    std::swap(upper, other.upper);
	    std::swap(pos, other.pos);
	    return *this;
	  }

	  iterator& operator=(const iterator& other) {
	    return *this = iterator(other);
	  }
	};
	iterator begin() {
	  return iterator(*upper, 0);
	}
	iterator end() {
	  return iterator(*upper, upper->size);
	}
      };

      template <typename Return,
		typename NonRefReturn = typename std::decay<Return>
		::type,
		typename = typename details::position<NonRefReturn, Types...>::type
		>
      random_container<NonRefReturn> get(selector<Return>) {
	return random_container<NonRefReturn>(*this);
      }

      template <typename Return,
		typename = typename
		std::enable_if<is_constructible<Return(const Return&)>::value ||
			       is_constructible<Return(Return&&)>::value>::type>
      Return& generate(int prob = 1) {
	static_assert(details::number_function_returns
		      <Return, Functions...>::value > 0,
		      "No function with such return");
	//FIXME: Check that some ground terms are accessible.

	std::vector<Return>& l =
	  std::get<details::position<Return, Types...>::value>(terms);
	if (!l.empty()) {
	  if (prob < 0) {
	    return l[std::uniform_int_distribution<size_t>
		     (0, l.size()-1)(g)];
	  }
	  if (prob > 0)
	    if (std::uniform_int_distribution<int>(0,prob)(g)) {
	      return l[std::uniform_int_distribution<size_t>
		       (0, l.size()-1)(g)];
	    }
	}

	size_t fun_to_call =
	  std::uniform_int_distribution<size_t>
	  (0, details::number_function_returns<Return,
					       Functions...>::value-1)(g);
	Return ret = call_with([this, fun_to_call] (Functions&...
						    funs) {
				 return this->call
				 (selector<Return>(),
				  fun_to_call,
				  funs...);
			       }, functions);

	l.push_back(std::move(ret));
	return l.back();
      }

      template <typename Return,
		typename = typename
		std::enable_if<!(is_constructible<Return(const Return&)>::value ||
				 is_constructible<Return(Return&&)>::value)>::type,
		typename = void>
      Return& generate(int = 1) {
	static_assert(details::number_function_returns
		      <Return, Functions...>::value > 0,
		      "No function with such return");
	size_t fun_to_call =
	  std::uniform_int_distribution<size_t>
	  (0, details::number_function_returns<Return&,
					       Functions...>::value-1)(g);

	return call_with([this, fun_to_call] (Functions&...
					      funs) -> Return& {
			   return this->call
			     (selector<Return&>(),
			      fun_to_call,
			      funs...);
			 }, functions);
      }

    };

    /// \brief Build a random term generator
    ///
    /// \param g A random generator engine.
    /// \param functions A list of functions.
    template <typename Generator, typename... Functions>
    generator<Generator, typename wrapped<Functions>::type...>
    operator()(Generator& g, Functions&&... functions) const {
      return generator<Generator, typename std::decay<typename wrapped<Functions>::type>::type...>
	(size, g, wrap(std::forward<Functions>(functions))...);
    }
  };

}

#endif
